/*--------------------------------------------------------------------
This source distribution is placed in the public domain by its author,
Jason Papadopoulos. You may use it for any purpose, free of charge,
without having to notify anyone. I disclaim any responsibility for any
errors.

Optionally, please be nice and tell me if you find this source to be
useful. Again optionally, if you add to the functionality present here
please consider making those additions public too, so that others may 
benefit from your work.	
       				   --jasonp@boo.net 9/24/08
--------------------------------------------------------------------*/

#include "poly.h"

#if MAX_POLY_DEGREE > 6
#error "polynomial degree > 6 not supported"
#endif

typedef struct {
	double power;
	dpoly_t rpoly;
	dpoly_t apoly;
} dparam_t;

/*------------------------------------------------------------------*/
static double get_polyval(dpoly_t *poly, double x, double h) {

	/* evaluate poly(x+h) using a taylor series centered
	   at poly(x). We compute poly(x) and the correction
	   from poly(x) separately, because in practice x and h
	   can have such different magnitudes that x+h could
	   equal x to machine precision. */

	double base; 
	double off;
	double hpow;
	double *p = poly->coeff;

	switch (poly->degree) {
	case 0:
		base = p[0];
		off = 0;
		break;
	case 1:
		base = p[1]*x+p[0];
		off = p[1]*h;
		break;
	case 2:
		base = (p[2]*x+p[1])*x+p[0];
		hpow = h;
		off = hpow * (2*p[2]*x+p[1]);
		hpow *= h;
		off += hpow * p[2];
		break;
	case 3:
		base = ((p[3]*x+p[2])*x+p[1])*x+p[0];
		hpow = h;
		off = hpow * ((3*p[3]*x+2*p[2])*x+p[1]);
		hpow *= h;
		off += hpow * (3*p[3]*x+p[2]);
		hpow *= h;
		off += hpow * p[3];
		break;
	case 4:
		base = (((p[4]*x+p[3])*x+p[2])*x+p[1])*x+p[0];
		hpow = h;
		off = hpow * (((4*p[4]*x+3*p[3])*x+2*p[2])*x+p[1]);
		hpow *= h;
		off += hpow * ((6*p[4]*x+3*p[3])*x+p[2]);
		hpow *= h;
		off += hpow * (4*p[4]*x+p[3]);
		hpow *= h;
		off += hpow * p[4];
		break;
	case 5:
		base = ((((p[5]*x+p[4])*x+p[3])*x+p[2])*x+p[1])*x+p[0];
		hpow = h;
		off = hpow * ((((5*p[5]*x+4*p[4])*x+3*p[3])*x+2*p[2])*x+p[1]);
		hpow *= h;
		off += hpow * (((10*p[5]*x+6*p[4])*x+3*p[3])*x+p[2]);
		hpow *= h;
		off += hpow * ((10*p[5]*x+4*p[4])*x+p[3]);
		hpow *= h;
		off += hpow * (5*p[5]*x+p[4]);
		hpow *= h;
		off += hpow * p[5];
		break;
	case 6:
		base = (((((p[6]*x+p[5])*x+p[4])*x+p[3])*x+p[2])*x+p[1])*x+p[0];
		hpow = h;
		off = hpow * (((((6*p[6]*x+5*p[5])*x+4*p[4])*x+
						3*p[3])*x+2*p[2])*x+p[1]);
		hpow *= h;
		off += hpow * ((((15*p[6]*x+10*p[5])*x+
						6*p[4])*x+3*p[3])*x+p[2]);
		hpow *= h;
		off += hpow * (((20*p[6]*x+10*p[5])*x+4*p[4])*x+p[3]);
		hpow *= h;
		off += hpow * ((15*p[6]*x+5*p[5])*x+p[4]);
		hpow *= h;
		off += hpow * (6*p[6]*x+p[5]);
		hpow *= h;
		off += hpow * p[6];
		break;
	default:
		base = off = 0;
		break;
	}
	return base + off;
}

/*------------------------------------------------------------------*/
static double integrand(double x, double h, void *params) {

	/* callback for numerical integration */

	dparam_t *aux = (dparam_t *)params;
	double polyval;

	polyval = get_polyval(&aux->rpoly, x, h) *
		  get_polyval(&aux->apoly, x, h);
	return pow(fabs(polyval), aux->power);
}

/*------------------------------------------------------------------*/
static int compare_complex(const void *x, const void *y) {
	complex_t *xx = (complex_t *)x;
	complex_t *yy = (complex_t *)y;

	if (xx->r > yy->r)
		return 1;
	if (xx->r < yy->r)
		return -1;
	return 0;
}

/*------------------------------------------------------------------*/
#define INTEGRATE_LIMIT 1e12

uint32 analyze_poly_size(integrate_t *integ_aux,
			dpoly_t *rpoly, dpoly_t *apoly, 
			double *result) {

	uint32 i;
	uint32 rdeg, adeg;
	dparam_t params;
	complex_t roots[2*MAX_POLY_DEGREE];
	uint32 num_roots;
	double endpoints[4*MAX_POLY_DEGREE];
	uint32 num_endpoints;
	double accum;
	double err;
	double curr_endpoint;

	rdeg = rpoly->degree;
	adeg = apoly->degree;
	params.power = -2.0 / (rdeg + adeg);
	params.rpoly = *rpoly;
	params.apoly = *apoly;

	/* Rather than use Murphy's method to rate the average
	   size of polynomial values, we use a result of D.J.
	   Bernstein: For rational poly R(x) and algebraic poly
	   A(x), the number of polynomial values where x is rational
	   and the size of A(x)*R(x) achieves a given value is 
	   proportional to the integral of

	   	dx / ((R(x) * A(x))^2) ^ (1/(deg(R(x))+deg(A(x))))

	   for x from -infinity to +infinity. Larger values of
	   this integral imply more values that get found by
	   a siever. Bernstein writes that this estimate is 
	   "extremely accurate", but has not elaborated to date.
	   The integration variable x refers to the a/b values used
	   by a siever, and it is both more realistic and simpler to
	   make the integration interval finite but large 
	   
	   The integrand takes on a wide range of values depending
	   on how close x is to a root of R(x)*A(x), so we have to
	   partition the integral */


	/* find the roots of R(x)*A(x) and sort in order of 
	   increasing real part */

	if (find_poly_roots(params.rpoly.coeff, rdeg, roots)) {
		printf("rational poly rootfinder failed\n");
		return 1;
	}
	if (find_poly_roots(params.apoly.coeff, adeg, roots + rdeg)) {
		printf("algebraic poly rootfinder failed\n");
		return 1;
	}
	qsort(roots, (size_t)(rdeg + adeg),
			sizeof(complex_t), compare_complex);

	/* squeeze out roots whose real part would not lie
	   in the integration interval, along with all complex 
	   conjugates of complex roots.

	   We cannot remove all complex roots because if the
	   imaginary part of the root is small then the value
	   of A(x)*R(x) is 'almost' a singularity at this point,
	   and ignoring it would throw off the numerical integrator */
	   
	for (i = num_roots = 0; i < rdeg + adeg; i++) {
		if (roots[i].i < 0.0 || 
		    roots[i].r <= -INTEGRATE_LIMIT ||
		    roots[i].r >= INTEGRATE_LIMIT)
			continue;
		roots[num_roots++] = roots[i];
	}

	/* convert the collection of roots into a collection
	   of integration endpoints */

	num_endpoints = 0;
	curr_endpoint = endpoints[num_endpoints++] = -INTEGRATE_LIMIT;
	for (i = 0; i <= num_roots; i++) {
		double new_endpoint;

		if (i == num_roots)
			new_endpoint = INTEGRATE_LIMIT;
		else
			new_endpoint = roots[i].r;

		/* if the integration interval is very large, add
		   intermediate intervals. Since the integrand varies 
		   widely in size over the interval, without these 
		   intermediate endpoints to separate regions of 
		   similar integrand size the integrator will do 
		   2-4x as much work as it has to (it's not adaptive) */

		if (new_endpoint - curr_endpoint > 1e8) {
			if (fabs(new_endpoint) < fabs(curr_endpoint)) {
				endpoints[num_endpoints++] = new_endpoint-1e6;
				endpoints[num_endpoints++] = new_endpoint-1e3;
			}
			else {
				endpoints[num_endpoints++] = curr_endpoint+1e3;
				endpoints[num_endpoints++] = curr_endpoint+1e6;
			}
		}
		endpoints[num_endpoints++] = new_endpoint;
		curr_endpoint = new_endpoint;
	}

	/* proceed from interval to interval, adding up the
	   values of the integral and saving the maximum error */

	curr_endpoint = endpoints[0];
	accum = err = 0.0;

	for (i = 1; i < num_endpoints; i++) {
		double new_endpoint = endpoints[i];

		integrate_run(integ_aux, integrand,
				&params, curr_endpoint, new_endpoint);

		accum += integ_aux->result;
		err = MAX(err, integ_aux->error);
		curr_endpoint = new_endpoint;
	}

	*result = accum;
	/* test for convergence */
	if (err > SIZE_EPS * fabs(accum)) {
		printf("integrator failed %e %e\n", 
				err, SIZE_EPS * fabs(accum));
		return 2;
	}
	return 0;
}

/*------------------------------------------------------------------*/
#define ROOT_BUF_SIZE 200

static double get_root_freq(mp_poly_t *poly, 
			uint32 initial_root, uint32 p) {
	
	/* find the number of roots of poly mod p^k for
	   all k where p^k <= PRIME_BOUND, then add up
	   the resulting contribution to random values of
	   poly */

	uint32 i, j;
	uint32 coeffs[MAX_POLY_DEGREE + 1];
	uint32 roots[ROOT_BUF_SIZE];
	uint32 new_roots[ROOT_BUF_SIZE];
	uint32 max_power, power, degree;
	uint32 num_roots, num_new_roots;
	double contrib;

	/* find the largest power of p that does not exceed 2^16 */

	max_power = 1;
	power = p;
	while (1) {
		uint32 next_power = power * p;
		if (next_power > 65000)
			break;

		power = next_power;
		max_power++;
	}

	/* compute poly mod power */

	degree = poly->degree;
	for (i = 0; i <= degree; i++) {
		j = mp_mod_1(&(poly->coeff[i].num), power);
		if (j > 0 && poly->coeff[i].sign == NEGATIVE)
			j = power - j;
		coeffs[i] = j;
	}
	while (degree > 0 && coeffs[degree] == 0)
		degree--;

	/* there is only one root mod p */

	num_roots = 1;
	roots[0] = initial_root;
	power = p;
	contrib = 1.0 / power;

	for (i = 2; i <= max_power; i++) {

		uint32 next_power = power * p;
		uint32 x;

		/* search for roots mod p^i using the roots
		   mod p^(i-1) */

		for (j = num_new_roots = 0; j < num_roots; j++) {

			/* all roots mod p^i must have a known value
			   mod p^(i-1). One root mod p^(i-1) can be a
			   multiple root, and can give rise to several 
			   roots mod p^i, or to no roots at all */

			for (x = roots[j]; x < next_power; x += power) {

				uint32 k;
				uint32 polyval = coeffs[degree] % next_power;
				for (k = degree; k; k--) {
					polyval = (polyval * x + 
						coeffs[k-1]) % next_power;
				}
				if (polyval == 0 && 
				    num_new_roots < ROOT_BUF_SIZE) {
					new_roots[num_new_roots++] = x;
				}
			}
		}

		if (num_new_roots == 0)
			break;

		/* add in the contribution due to roots mod p^i 
		   and make the new roots into the old roots */

		contrib += (double)num_new_roots / next_power;
		memcpy(roots, new_roots, num_new_roots * sizeof(uint32));
		power = next_power;
		num_roots = num_new_roots;
	}

	/* for ordinary roots, contrib would be 1/(p-1)
	   asymptotically */

	return contrib * p / (p + 1);
}

/*------------------------------------------------------------------*/
#define SMALL_PRIME_BOUND 100

uint32 analyze_poly_roots(mp_poly_t *poly, uint32 prime_bound,
				double *result) {

	/* analyze a polynomial for root properties (i.e.
	   compute Murphy's 'alpha' value) */

	uint32 i, j;
	double root_score;
	uint32 prime;
	mp_poly_t rev_poly;

	/* all linear polynomials have the same root properties */

	if (poly->degree == 1) {
		*result = 0.0;
		return 0;
	}

	/* handling projective roots requires a reversed version
	   of poly (we find roots of poly(1/x) instead of poly(x)) */

	memset(&rev_poly, 0, sizeof(mp_poly_t));
	j = poly->degree;
	for (i = 0; i <= j; i++) {
		signed_mp_copy(poly->coeff + j - i, rev_poly.coeff + i);
	}
	while (j && mp_is_zero(&rev_poly.coeff[j].num)) {
		j--;
	}
	rev_poly.degree = j;

	/* roots of (poly mod p) can be either ordinary or special.
	   An ordinary root mod p corresponds to exactly one root
	   mod p^k for all k > 1. Murphy shows that for each ordinary 
	   root of (poly mod p) the contribution to the size of a
	   random sieve value due to all powers of p is log(p) * 
	   p/(p^2-1). 
	   
	   We compare this to the contribution of powers of p to 
	   the size of a random number. 1 in p random numbers are 
	   divisible by p, 1 in p^2 random numbers are divisible by
	   p^2, etc. Adding up all of these contributions and summing
	   the resulting geometric series yields an average contribution
	   of log(p) / (p-1). 
	   
	   Hence the root score for poly is the sum of the difference 
	   between the contribution of powers of p in sieve values 
	   compared to the contribution to a random number. When there 
	   are k ordinary roots mod p, root_score is the sum of 

	               log(p) * (1/(p-1) - k*p/(p^2-1))

	   for the first few p, meaning that dividing the small primes 
	   out of sieve values makes their size exp(root_score) times 
	   smaller than dividing the small primes out of a random number.

	   When the number of roots mod p^k changes with k, Murphy's 
	   formula for the contribution is inaccurate (usually it's too
	   small), so we have to take special measures to compute the 
	   contribution of the roots accurately. Roots become special 
	   if p divides the discriminant of poly, and basically the only 
	   effective way to compute the size of the contribution from 
	   special roots is to find the number of roots mod p^k for the
	   first few k and then explicitly sum the first few terms in 
	   the (formerly geometric) series. Many thanks to Kleinjung 
	   and Franke whose pol5 code handles special roots efficiently */

	root_score = 0;
	for (i = prime = 0; i < PRECOMPUTED_NUM_PRIMES; i++) {

		uint32 roots[MAX_POLY_DEGREE];
		uint32 mult[MAX_POLY_DEGREE];
		uint32 high_coeff;
		uint32 num_roots, num_ordinary_roots;
		double contrib;

		/* test if next prime will exceed the bound */

		if (prime + prime_delta[i] > SMALL_PRIME_BOUND || 
		    prime + prime_delta[i] > prime_bound)
			break;

		/* find the roots of poly mod prime and determine if
		   any are multiple roots (this is the telltale sign
		   that the root is special) */

		prime += prime_delta[i];
		num_roots = poly_get_zeros_and_mult(roots, mult, poly, 
					prime, &high_coeff);

		contrib = 1.0 / (prime - 1);

		/* test each root for specialness */

		for (j = num_ordinary_roots = 0; j < num_roots; j++) {
			if (mult[j] == 0)
				num_ordinary_roots++;
			else
				contrib -= get_root_freq(poly, roots[j], prime);
		}
		
		/* projective roots are always handled specially */

		if (high_coeff == 0)
			contrib -= get_root_freq(&rev_poly, 0, prime);

		/* add in the root score contribution for prime */

		root_score += log((double)prime) * (contrib - 
						(double)num_ordinary_roots * 
						prime / (prime * prime - 1));
	}

	/* for the larger primes, assume that all roots are
	   ordinary, so that only the number of roots modulo
	   prime i matter */

	for (; i < PRECOMPUTED_NUM_PRIMES; i++) {
		uint32 num_roots;
		uint32 dummy_roots[MAX_POLY_DEGREE];
		uint32 high_coeff;

		prime += prime_delta[i];
		if (prime > prime_bound)
			break;

		num_roots = poly_get_zeros(dummy_roots, poly, 
					prime, &high_coeff, 1);
		if (high_coeff == 0)
			num_roots++;

		root_score += (1.0 - (double)num_roots * prime / 
					(prime + 1)) * log((double)prime) / 
					(prime - 1);
	}

	*result = root_score;
	return 0;
}

/*------------------------------------------------------------------*/
uint32 analyze_poly_roots_projective(mp_poly_t *poly, 
				uint32 prime_bound,
				double *result) {

	/* compute the contribution of projective roots
	   to random vaues of poly() */

	uint32 i, j;
	double root_score;
	uint32 prime;
	mp_poly_t rev_poly;

	/* handling projective roots requires a reversed version
	   of poly (we find roots of poly(1/x) instead of poly(x)) */

	memset(&rev_poly, 0, sizeof(mp_poly_t));
	j = poly->degree;
	for (i = 0; i <= j; i++) {
		signed_mp_copy(poly->coeff + j - i, rev_poly.coeff + i);
	}
	while (j && mp_is_zero(&rev_poly.coeff[j].num)) {
		j--;
	}
	rev_poly.degree = j;

	root_score = 0;
	for (i = prime = 0; i < PRECOMPUTED_NUM_PRIMES; i++) {

		prime += prime_delta[i];
		if (prime > prime_bound)
			break;

		/* only projective roots contribute */

		if (mp_mod_1(&poly->coeff[poly->degree].num, prime) == 0) {
			root_score -= log((double)prime) *
					get_root_freq(&rev_poly, 0, prime);
		}
	}

	*result = root_score;
	return 0;
}

/*------------------------------------------------------------------*/
void get_poly_combined_score(poly_select_t *poly) {

	/* An interesting problem: you have a root score and
	   a size score for each polynomial; how do you combine
	   them to get an overall rating? The value of root_score
	   implies that values of R(x)*A(x) effectively have size
	   R(x)*A(x)*exp(root_score) after dividing out the small
	   primes from sieve values. Technically this will
	   overestimate the size, because root_score includes a 
	   bias that accounts for random numbers; but this bias
	   is the same for all polynomials. The exponential term 
	   is constant, so pulling this expression out of the 
	   integrand used to calculate the size above gives: */

	poly->combined_score = poly->size_score * 
				pow(exp(poly->root_score), 
					-2.0 / (poly->rpoly.degree + 
						poly->apoly.degree));
}

/*------------------------------------------------------------------*/
void save_poly(poly_config_t *config, poly_select_t *poly) {

	/* save the polynomial if its combined score is the 
	   best seen so far. A really thorough implementation
	   would do this in two stages: the best X polynomials
	   would all be kept in stage 1, and in stage 2 the 
	   complete list (possibly combined from the output of
	   different machines) would get a more accurate 
	   root_score computed for each polynomial. The top
	   few such polynomials would then be subjected to 
	   sieving experiments */

	if (poly->combined_score > config->heap[0]->combined_score) {
		memcpy(config->heap[0], poly, sizeof(poly_select_t));
	}
}

/*------------------------------------------------------------------*/
void analyze_poly(poly_config_t *config, poly_select_t *poly) {

	/* analyze a polynomial for sieving goodness
	  
	   The analysis routines are general enough so that
	   any polynomials can be tested, independent of
	   degree and skewness. The score assigned is
	   directly comparable to that of any other polynomials 
	   given to this routine */

	uint32 i;
	double root_score_r, root_score_a;
	dpoly_t dr, da;
	mp_poly_t *rpoly = &poly->rpoly;
	mp_poly_t *apoly = &poly->apoly;

	poly->size_score = 0.0;
	poly->root_score = 0.0;
	poly->combined_score = 0.0;

	/* convert the polynomial coefficients from arbitrary
	   precision to double precision floating point */

	dr.degree = rpoly->degree;
	for (i = 0; i <= rpoly->degree; i++)
		dr.coeff[i] = mp_signed_mp2d(rpoly->coeff + i);

	da.degree = apoly->degree;
	for (i = 0; i <= apoly->degree; i++)
		da.coeff[i] = mp_signed_mp2d(apoly->coeff + i);

	if (analyze_poly_size(&config->integ_aux, 
				&dr, &da, &poly->size_score))
		return;

	if (analyze_poly_roots(&poly->rpoly, PRIME_BOUND, &root_score_r))
		return;
	if (analyze_poly_roots(&poly->apoly, PRIME_BOUND, &root_score_a))
		return;
	poly->root_score = root_score_r + root_score_a;

	get_poly_combined_score(poly);
}

/*------------------------------------------------------------------*/
void analyze_one_poly(msieve_obj *obj,
		      mp_poly_t *rpoly, mp_poly_t *apoly) {

	uint32 i;
	poly_config_t config;
	poly_select_t s;

	poly_config_init(obj, &config);
	memcpy(&s.rpoly, rpoly, sizeof(mp_poly_t));
	memcpy(&s.apoly, apoly, sizeof(mp_poly_t));
	analyze_poly(&config, &s);

	for (i = 0; i <= rpoly->degree; i++) {
		logprintf(obj, "R%u: %c%s\n", i,
			(rpoly->coeff[i].sign == POSITIVE? ' ' : '-'),
			mp_sprintf(&rpoly->coeff[i].num, 10,
						obj->mp_sprintf_buf));
	}
	for (i = 0; i <= apoly->degree; i++) {
		logprintf(obj, "A%u: %c%s\n", i,
			(apoly->coeff[i].sign == POSITIVE? ' ' : '-'),
			mp_sprintf(&apoly->coeff[i].num, 10,
						obj->mp_sprintf_buf));
	}
	logprintf(obj, "size score = %e, Murphy alpha = %f, combined = %e\n",
			s.size_score, s.root_score, s.combined_score);
	poly_config_free(&config);
}

/*------------------------------------------------------------------*/
void poly_config_init(msieve_obj *obj, poly_config_t *config) {

	/* one-time initialization for polynomial search */

	uint32 i;
	char buf[256];

	config->heap_num_filled = 0;
	for (i = 0; i < POLY_HEAP_SIZE; i++) {
		config->heap[i] = (poly_select_t *)xcalloc((size_t)1, 
					sizeof(poly_select_t));
	}

	integrate_init(&config->integ_aux, 1e-100, 0.01 * SIZE_EPS,
			double_exponential, 8);

	sprintf(buf, "%s.p", obj->savefile.name);
	config->all_poly_file = fopen(buf, "a");
	if (config->all_poly_file == NULL) {
		printf("error: cannot open all-poly file\n");
		exit(-1);
	}
}

/*------------------------------------------------------------------*/
void poly_config_free(poly_config_t *config) {

	uint32 i;

	fclose(config->all_poly_file);
	for (i = 0; i < POLY_HEAP_SIZE; i++)
		free(config->heap[i]);
	integrate_free(&config->integ_aux);
}
