#include "stage1.h"

static const uint32 pr_mod5[NPR5] = {
	11, 31, 41, 61, 71, 101, 131, 151, 181, 191,
	211, 241, 251, 271, 281, 311, 331, 401, 421, 431,
	461, 491, 521, 541, 571, 601, 631, 641, 661, 691,
	701, 751, 761, 811, 821, 881, 911, 941, 971, 991,
	1021, 1031, 1051, 1061, 1091, 1151, 1171, 1181, 1201, 1231
};				/* primes =1 mod 5, < 2^(31/3) */

static const unsigned char ucmask[8] = { 
	0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80 
};

/*------------------------------------------------------------------------*/
void
stage1_stat_init(stage1_stat_t *stats)
{
	memset(stats, 0, sizeof(stage1_stat_t));
}

/*------------------------------------------------------------------------*/
void
stage1_stat_free(stage1_stat_t *stats)
{
	memset(stats, 0, sizeof(stage1_stat_t));
}

/*------------------------------------------------------------------------*/
void 
stage1_bounds_init(bounds_t *bounds, poly_stage1_t *data)
{
	mpz_init_set(bounds->gmp_a5_begin, data->gmp_a5_begin);
	mpz_init_set(bounds->gmp_a5_end, data->gmp_a5_end);
	bounds->norm_max = data->norm_max;
	bounds->p0_limit = data->p0_limit;
	if (bounds->p0_limit == 0 || bounds->p0_limit > P0_MAX)
		bounds->p0_limit = P0_MAX;
}

/*------------------------------------------------------------------------*/
void 
stage1_bounds_free(bounds_t *bounds)
{
	mpz_clear(bounds->gmp_a5_begin);
	mpz_clear(bounds->gmp_a5_end);
}

/*------------------------------------------------------------------------*/
void 
stage1_bounds_update(bounds_t *bounds, double N, double a5)
{
	bounds->skewness_max = pow(bounds->norm_max / a5, 0.4);
	bounds->skewness_min = pow(N / a5 / 
				pow(bounds->norm_max, 5.), 2./15.);
	bounds->a3_max = bounds->norm_max / 
				sqrt(bounds->skewness_min);
	bounds->p_size_max = log(bounds->a3_max / 
				bounds->skewness_min);
}

/*------------------------------------------------------------------------*/
void
curr_poly_init(curr_poly_t *curr, bounds_t *bounds, poly_stage1_t *data)
{
	int i;
	double dbl_a5, dbl_N, dbl_a5_min, dbl_a5_max;

	mpz_init_set(curr->gmp_N, data->gmp_N);
	mpz_init(curr->gmp_root);
	mpz_init(curr->gmp_prod);
	mpz_init(curr->gmp_prod_base);
	mpz_init(curr->gmp_help1);
	mpz_init(curr->gmp_help2);
	mpz_init(curr->gmp_help3);
	mpz_init(curr->gmp_help4);
	mpz_init(curr->gmp_a5);
	mpz_init(curr->gmp_d);
	mpz_init(curr->gmp_m0);
	curr->callback = data->callback;
	curr->callback_data = data->callback_data;

	mpz_fdiv_q_ui(curr->gmp_a5, bounds->gmp_a5_begin, MULTIPLIER);
	mpz_mul_ui(curr->gmp_a5, curr->gmp_a5, MULTIPLIER);
	if (mpz_cmp(curr->gmp_a5, bounds->gmp_a5_begin) < 0)
		mpz_add_ui(curr->gmp_a5, curr->gmp_a5, MULTIPLIER);

	dbl_N = mpz_get_d(curr->gmp_N);
	dbl_a5 = mpz_get_d(curr->gmp_a5);

	dbl_a5_max = pow(bounds->norm_max, 8.) / dbl_N;
	dbl_a5_max = sqrt(dbl_a5_max);
	dbl_a5_min = 0.;
	for (i = 0; i < data->npr_in_p; i++)
		dbl_a5_min += log((double) pr_mod5[i]);
	dbl_a5_min *= 5.;
	dbl_a5_min -= 10. * log(bounds->norm_max);
	dbl_a5_min += log(dbl_N);
	dbl_a5_min = exp(dbl_a5_min);
	dbl_a5_min = floor(dbl_a5_min) + 1.;

	if (dbl_a5_min > dbl_a5) {
		printf("a5_begin smaller than %f\n", dbl_a5_min);
		exit(-1);
	}
	if (dbl_a5_max < mpz_get_d(bounds->gmp_a5_end)) {
		printf("a5_end bigger than %f\n", dbl_a5_max);
		exit(-1);
	}

	mpz_sub_ui(curr->gmp_a5, curr->gmp_a5, MULTIPLIER);
	printf("%.0lf %.0lf\n", mpz_get_d(curr->gmp_a5), 
			mpz_get_d(bounds->gmp_a5_end));

}

/*------------------------------------------------------------------------*/
void
curr_poly_free(curr_poly_t *curr)
{
	mpz_clear(curr->gmp_N);
	mpz_clear(curr->gmp_prod);
	mpz_clear(curr->gmp_prod_base);
	mpz_clear(curr->gmp_root);
	mpz_clear(curr->gmp_help1);
	mpz_clear(curr->gmp_help2);
	mpz_clear(curr->gmp_help3);
	mpz_clear(curr->gmp_help4);
	mpz_clear(curr->gmp_a5);
	mpz_clear(curr->gmp_d);
	mpz_clear(curr->gmp_m0);
}

/*------------------------------------------------------------------------*/
static uint32
is_5power(uint32 a, uint32 p)
{			/* p!=5, 0 is not considered as fifth power */
	if (p % 5 != 1)
		return 0;
	if (mp_expo_1(a, (p - 1) / 5, p) == 1)
		return 1;
	return 0;
}

void
rat_factor_init(rat_factor_t *rat_factors, mpz_t N, 
		mpz_t a5_start, uint32 npr_in_p)
{
	uint32 i, j;
	uint32 len;
	rat_prime_t *primes = rat_factors->primes;

	for (i = 0; i < NPR5; i++) {
		rat_prime_t *r = primes + i;
		uint32 p = pr_mod5[i];

		r->prime = p;
		r->logprime = log((double)p);

		r->N_mod_p2 = mpz_fdiv_ui(N, p * p);
		r->N_inv_mod_p = mp_modinv_1(mpz_fdiv_ui(N, p), p);
		r->sieve_step = mp_modmul_1(MULTIPLIER, r->N_inv_mod_p, p);
		r->sieve_start = mp_modmul_1(mpz_fdiv_ui(a5_start, p),
					    r->N_inv_mod_p, p);
	}

	for (i = 0; i < NPR5; i++) {
		for (j = 0; j < NPR5; j++) {
			if (i != j) {
				uint32 pi = primes[i].prime;
				uint32 pj = primes[j].prime;
				rat_factors->inv_table[i][j] =
						mp_modinv_1(pi % pj, pj);
			}
		}
	}

	for (i = len = 0; i < NPR5; i++)
		len += (pr_mod5[i] / 8 + 1);
	rat_factors->bitarrays[0] = (unsigned char *)xcalloc(len,
						sizeof(unsigned char));
	for (i = 1; i < NPR5; i++) {
		rat_factors->bitarrays[i] = rat_factors->bitarrays[i - 1] + 
						(pr_mod5[i - 1] / 8 + 1);
	}
	for (i = 0; i < NPR5; i++) {
		uint32 p = pr_mod5[i];
		unsigned char *bitarray = rat_factors->bitarrays[i];

		for (j = 1; j < p; j++) {
			if (is_5power(j, p)) {
				bitarray[j / 8] |= ucmask[j % 8];
			}
		}
	}

	rat_factors->npr_in_p = npr_in_p;
	rat_factors->p0_list_len_max = 256;
	rat_factors->p0_list = (aux_factor_t *)xmalloc(
					rat_factors->p0_list_len_max *
				       	sizeof(aux_factor_t));
}

/*------------------------------------------------------------------------*/
void
rat_factor_free(rat_factor_t *rat_factors)
{
	free(rat_factors->bitarrays[0]);
	free(rat_factors->p0_list);
}

/*------------------------------------------------------------------------*/
static void
fifth_root(uint32 *ro, uint32 a, uint32 p)
{
	uint32 ind = 0;
	uint32 r;

	for (r = 1; r < p; r++) {
		if (mp_expo_1(r, 5, p) == a) {
			ro[ind++] = r;
			if (ind == 5)
				break;
		}
	}
}

uint32
rat_factor_find(rat_factor_t *rat_factors, 
		mpz_t a5, double p_size_max)
{
	int i, j;
	uint32 h, p;
	double size;
	uint32 npr_in_p = rat_factors->npr_in_p;
	uint32 *a5_p_idx = rat_factors->a5_p_idx;

	for (i = j = 0; i < NPR5; i++) {
		rat_prime_t *r = rat_factors->primes + i;
		unsigned char *bitarray = rat_factors->bitarrays[i];

		h = r->sieve_start = mp_modadd_1(r->sieve_start, 
						r->sieve_step, r->prime);
		if (bitarray[h / 8] & ucmask[h % 8]) {
			r->my_idx = j;
			a5_p_idx[j] = i;
			j++;
		}
	}

	if (j < npr_in_p)
		return 0;
	for (i = 0, size = 0.0; i < npr_in_p; i++) {
		rat_prime_t *r = rat_factors->primes + a5_p_idx[i];
		size += r->logprime;
	}
	if (size > p_size_max)
		return 0;

	for (i = 0; i < j; i++) {
		rat_prime_t *r = rat_factors->primes + a5_p_idx[i];

		p = r->prime;
		r->a5_mod_p2 = mpz_fdiv_ui(a5, p * p);
		r->minus5a5_mod_p = mp_modmul_1(p - 5, 
						r->a5_mod_p2, p);
		h = r->a5_mod_p2 % p;
		h = mp_modmul_1(mp_modinv_1(h, p), 
				r->N_mod_p2, p);
		fifth_root(r->roots, h, p);
	}
	rat_factors->npr_total = j;
	return 1;
}

/*------------------------------------------------------------------------*/
static void
aux_factor_store(rat_factor_t *r,
		unsigned int n, unsigned int root)
{
	uint32 len = r->p0_list_len;

	if (len >= r->p0_list_len_max) {
		r->p0_list_len_max *= 2;
		r->p0_list = (aux_factor_t *) xrealloc(r->p0_list,
						  r->p0_list_len_max *
						  sizeof(aux_factor_t));
	}
	r->p0_list[len].p = n;
	r->p0_list[len].r = root;
	r->p0_list_len++;
}

/*------------------------------------------------------------------------*/
static unsigned int
aux_factor_phi(unsigned int n)
{				/* n<=2^16 */
	unsigned int i;
	unsigned int phi, d, p;

	if (n > 65536) {
		printf("aux_factor_phi\n");
		exit(-1);
	}
	d = n;
	phi = 1;
	for (i = p = 0; p < 256; i++) {
		p += prime_delta[i];
		if (d % p == 0) {
			phi *= (p - 1);
			d /= p;
			while (d % p == 0) {
				phi *= p;
				d /= p;
			}
			if (d == 1)
				break;
		}
	}
	if (d > 1)
		phi *= (d - 1);
	return phi;
}

/*------------------------------------------------------------------------*/
static void
aux_factor_roots(rat_factor_t *rat_factors, 
		unsigned int n, mpz_t gmp_N, mpz_t gmp_a5)
{
	int i;
	unsigned int h, r, a5, N, p;

	a5 = mpz_fdiv_ui(gmp_a5, n);
	if (mp_gcd_1(a5, n) != 1)
		return;
	N = mpz_fdiv_ui(gmp_N, n);
	if (n < 20) {
		if (n == 11)
			return;
		for (r = 1; r < n; r++) {
			h = mp_expo_1(r, 5, n);
			h = mp_modmul_1(h, a5, n);
			if (h == N)
				aux_factor_store(rat_factors, n, r);
		}
		return;
	}

	for (i = 0; i < NPR5; i++) {
		p = pr_mod5[i];
		if (p > n)
			break;
		if (n % p == 0)
			return;
	}

	/* for deg=5 exactly one root exists */
	h = aux_factor_phi(n);
	if (h % 5 == 0)
		return;

	r = 1 + h;
	while (r % 5)
		r += h;
	r /= 5;

	h = mp_modinv_1(a5, n);
	h = mp_modmul_1(h, N, n);
	r = mp_expo_1(h, r, n);
	aux_factor_store(rat_factors, n, r);
}

/*------------------------------------------------------------------------*/
void
rat_factor_fill_aux(rat_factor_t *rat_factors, 
			mpz_t N, mpz_t a5,
			double dbl_p0_max, uint32 p0_limit)
{
	uint32 n, p0m;

	rat_factors->p0_list[0].p = 1;
	rat_factors->p0_list[0].r = 0;
	rat_factors->p0_list_len = 1;
	if (dbl_p0_max > (double) p0_limit)
		p0m = p0_limit;
	else
		p0m = (uint32) rint(dbl_p0_max);

	if (p0m < 2)
		return;
	for (n = 2; n <= p0m; n++) {
		if (mp_gcd_1(n, MULTIPLIER) == 1)
			aux_factor_roots(rat_factors, n, N, a5);
	}
}

/*------------------------------------------------------------------------*/
static int
five_pow(int n)
{
	int res, i;

	res = 1;
	for (i = 0; i < n; i++)
		res *= 5;
	return res;
}

void
knapsack_data_init(knapsack_data_t *knap, uint32 npr_in_p)
{
	int i, j;
	int len11, len12, len21, len22;
	int s11len, s12len, s21len, s22len;

	memset(knap, 0, sizeof(knapsack_data_t));

	mpz_init(knap->gmp_disp);
	mpz_init(knap->gmp_D0);

	for (i = 0; i < NPR5; i++) {
		for (j = 0; j < 5; j++) {
			mpz_init(knap->gmp_D[i][j]);
		}
	}
	mpz_init(knap->gmp_kappa0);
	for (i = 0; i < NPR5; i++) {
		for (j = 0; j < 5; j++) {
			mpz_init(knap->gmp_kappa[i][j]);
		}
	}
	for (i = 0; i < NPR5 + 1; i++) {
		mpz_init(knap->gmp_kappa_help[i]);
	}

	knap->len1 = npr_in_p / 2;
	knap->len2 = npr_in_p - knap->len1;	/* len1<=len2 */
	len12 = knap->len1 / 2;
	len11 = knap->len1 - len12;
	len22 = knap->len2 / 2;
	len21 = knap->len2 - len22;
	s11len = five_pow(len11);
	s12len = five_pow(len12);
	s21len = five_pow(len21);
	s22len = five_pow(len22);
	knap->s1len = s11len * s12len;
	knap->s2len = s21len * s22len;
	knap->s1 = (uint64_t *) xmalloc(knap->s1len * sizeof(uint64_t));
	knap->s2 = (uint64_t *) xmalloc(knap->s2len * sizeof(uint64_t));
	knap->s1sort = (uint64_t *) xmalloc(knap->s1len * sizeof(uint64_t));
	knap->s2sort = (uint64_t *) xmalloc(knap->s2len * sizeof(uint64_t));
	knap->sort = (int *) xmalloc(NHASH * sizeof(int));
	knap->hashptr = (uint64_t **) xmalloc(NHASH * sizeof(uint64_t *));
	knap->stored_pairs = NULL;
}

/*------------------------------------------------------------------------*/
void
knapsack_data_free(knapsack_data_t *knap)
{
	int i, j;

	mpz_clear(knap->gmp_disp);
	mpz_clear(knap->gmp_D0);

	for (i = 0; i < NPR5; i++) {
		for (j = 0; j < 5; j++) {
			mpz_clear(knap->gmp_D[i][j]);
		}
	}
	mpz_clear(knap->gmp_kappa0);
	for (i = 0; i < NPR5; i++) {
		for (j = 0; j < 5; j++) {
			mpz_clear(knap->gmp_kappa[i][j]);
		}
	}
	for (i = 0; i < NPR5 + 1; i++) {
		mpz_clear(knap->gmp_kappa_help[i]);
	}

	free(knap->s1);
	free(knap->s2);
	free(knap->s1sort);
	free(knap->s2sort);
	free(knap->sort);
	free(knap->hashptr);
	free(knap->stored_pairs);
}

